#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <dirent.h>
#include <pwd.h>
#include <rsbac/types.h>
#include <rsbac/getname.h>
#include <rsbac/acl_getname.h>
#include <rsbac/syscalls.h>
#include <rsbac/error.h>
#include <rsbac/helpers.h>

#define MASK_PROG "acl_mask"

int verbose=0;
int recurse=0;
int printall=0;
int backup=0;
union rsbac_attribute_value_t value;
enum rsbac_target_t target;
enum rsbac_attribute_t attr;
char * progname;
rsbac_acl_rights_vector_t rights_vector;
boolean set = FALSE;

int process(char * name)
  {
    int res = 0;
    char tmp1[120];
    enum rsbac_target_t type = T_NONE;
    char type_n[5] = "";
    struct stat buf;
    struct rsbac_acl_syscall_arg_t   arg;
    struct rsbac_acl_syscall_n_arg_t arg_n;

    if(!strcmp(name,":DEFAULT:"))
      {
        if(target == T_FD)
          type = T_FILE;
        else
          type = target;
        switch(type)
          {
            case T_FILE:
              strcpy(type_n,"FILE");
              break;
            case T_DIR:
              strcpy(type_n,"DIR");
              break;
            case T_DEV:
              strcpy(type_n,"DEV");
              break;
            case T_IPC:
              strcpy(type_n,"IPC");
              arg.tid.ipc.type = I_none;
              break;
            case T_SCD:
              strcpy(type_n,"SCD");
              arg.tid.scd = AST_none;
              break;
            case T_USER:
              strcpy(type_n,"USER");
              arg.tid.user = RSBAC_NO_USER;
              break;
            case T_PROCESS:
              strcpy(type_n,"PROCESS");
              arg.tid.process = 0;
              break;
            default:
              fprintf(stderr, "Internal error: invalid target %u!\n",
                      target);
              return(1);
          }
        arg_n.name = NULL;
      }
    else
      {
        switch(target)
          {
            case T_FILE:
            case T_DIR:
            case T_FD:
            case T_DEV:
              if (lstat(name,&buf))
                {
                  fprintf(stderr, "stat for file %s returned error: %s\n",
                          name,
                          strerror(errno));
                  return(1);
                }
            default:
              break;
          }
        arg_n.name = name;
        switch(target)
          {
            case T_FILE:
              if(   (S_ISREG(buf.st_mode))
                 || (S_ISBLK(buf.st_mode))
                 || (S_ISCHR(buf.st_mode)) )
                {
                  type = T_FILE;
                  strcpy(type_n,"FILE");
                }
              else
                {
                  if(verbose)
                    printf("%s is no file, skipped\n",
                           name);
                  goto do_recurse;
                }
              break;
            case T_DIR:
              if(S_ISDIR(buf.st_mode))
                {
                  type = T_DIR;
                  strcpy(type_n,"DIR");
                }
              else
                {
                  if(verbose)
                    printf("%s is no dir, skipped\n",
                           name);
                  goto do_recurse;
                }
              break;
            case T_FD:
              if(S_ISDIR(buf.st_mode))
                {
                  type = T_DIR;
                  strcpy(type_n,"DIR");
                }
              else
              if(   (S_ISREG(buf.st_mode))
                 || (S_ISBLK(buf.st_mode))
                 || (S_ISCHR(buf.st_mode)) )
                {
                  type = T_FILE;
                  strcpy(type_n,"FILE");
                }
              else
                {
                  if(verbose)
                    printf("%s is neither file nor dir, skipped\n",
                           name);
                  goto do_recurse;
                }
              break;
            case T_DEV:
              if(   (S_ISBLK(buf.st_mode))
                 || (S_ISCHR(buf.st_mode)) )
                {
                  type = T_DEV;
                  strcpy(type_n,"DEV");
                }
              else
                {
                  if(verbose)
                    printf("%s is no file, skipped\n",
                           name);
                  goto do_recurse;
                }
              break;
            case T_SCD:
              arg.tid.scd = get_acl_scd_type_nr(name);
              if((arg.tid.scd == ST_none) || (arg.tid.scd == AST_none))
                {
                  fprintf(stderr, "%s is no valid SCD name, skipped\n",
                           name);
                  return(1);
                }
              break;

            default:
              fprintf(stderr, "Invalid target %u for %s, skipped!\n",
                      target, name);
              return(1);
          }
      }
    if(verbose)
      printf("Processing %s '%s'\n",
             type_n,
             name);
    
    if(set)
      {
        switch(target)
          {
            case T_FILE:
            case T_DIR:
            case T_FD:
            case T_DEV:
              arg_n.target = type;
              arg_n.rights = rights_vector;
              res = rsbac_acl_n(ACLC_set_mask, &arg_n);
              break;

            default:
              arg.target = type;
              arg.rights = rights_vector;
              res = rsbac_acl(ACLC_set_mask, &arg);
          }
      }
    else
      {
        switch(target)
          {
            case T_FILE:
            case T_DIR:
            case T_FD:
            case T_DEV:
              res = rsbac_acl_get_mask_n(type, name, &rights_vector);
              break;

            default:
              res = rsbac_acl_get_mask(target, &arg.tid, &rights_vector);
          }
      }
    if(res)
      {
        get_error_name(tmp1,res);
        fprintf(stderr, "%s: error: %s\n", name, tmp1);
        goto do_recurse;
      }
    if(!set)
      {
        if(backup)
          {
            if(rights_vector != RSBAC_ACL_DEFAULT_MASK)
              {
                printf("%s -sv",
                       MASK_PROG);
                if(printall)
                  {
                    int i;

                    for (i=0; i<R_NONE; i++)
                      if(rights_vector & ((rsbac_acl_rights_vector_t) 1 << i))
                        printf(" %s", get_request_name(tmp1,i));
                    for (i=RSBAC_ACL_SPECIAL_RIGHT_BASE; i<ACLR_NONE; i++)
                      if(rights_vector & ((rsbac_acl_rights_vector_t) 1 << i))
                        printf(" %s", get_acl_special_right_name(tmp1,i));
                  }
                else
                  {
                    printf(" %s", u64tostracl(tmp1, rights_vector));
                  }
                printf(" %s %s\n", type_n, name);
              }
          }
        else /* no backup */
          {
            printf("%s: %s\n",
                   name,
                   u64tostracl(tmp1, rights_vector));
            if(printall)
              {
                int i;

                for (i=0; i<R_NONE; i++)
                  if(rights_vector & ((rsbac_acl_rights_vector_t) 1 << i))
                    printf("  %s\n", get_request_name(tmp1,i));
                for (i=RSBAC_ACL_SPECIAL_RIGHT_BASE; i<ACLR_NONE; i++)
                  if(rights_vector & ((rsbac_acl_rights_vector_t) 1 << i))
                    printf("  %s\n", get_acl_special_right_name(tmp1,i));
              }
          }
      }

do_recurse:
    if (   (type == T_DIR)
        && recurse)
      {
        DIR * dir_stream_p;
        struct dirent * dirent_p;
        char name2[512];

        if(S_ISLNK(buf.st_mode))
          return(0);
        if(!(dir_stream_p = opendir(name)))
          {
            fprintf(stderr, "opendir for dir %s returned error: %s\n",
                   name,
                   strerror(errno));
            return(-2);
          }
        while((dirent_p = readdir(dir_stream_p)))
          {
            if(   (strcmp(".",dirent_p->d_name))
               && (strcmp("..",dirent_p->d_name)) )
              {
                strcpy(name2,name);
                strcat(name2,"/");
                strcat(name2,dirent_p->d_name);
                process(name2);
              }
          }
        closedir(dir_stream_p);
      }
    return(0);
  }

int main(int argc, char ** argv)
{
  int res = 0;
  int i;
  enum  rsbac_acl_special_rights_t right;
  boolean rused = FALSE;
  boolean wused = FALSE;
  
  progname = argv[0];
  while((argc > 1) && (argv[1][0] == '-'))
    {
      char * pos = argv[1];
      pos++;
      while(*pos)
        {
          switch(*pos)
            {
              case 'v':
                verbose++;
                break;
              case 'r':
                recurse=1;
                break;
              case 's':
                set = TRUE;
                break;
              case 'p':
                printall=1;
                break;
              case 'b':
                backup=1;
                break;
              case 'n':
                {
                  char tmp[80];
                  for(i=0; i<ST_none; i++)
                    printf("%s\n", get_scd_type_name(tmp, i));
                  for(i=AST_min; i<AST_none; i++)
                    printf("%s\n", get_acl_scd_type_name(tmp, i));
                  exit(0);
                }
              default:
                fprintf(stderr, "%s: unknown parameter %c\n", progname, *pos);
            }
          pos++;
        }
      argv++;
      argc--;
    }
  if (argc > 1)
    {
      while(argc > 2)
        {
          if(strlen(argv[1]) == ACLR_NONE)
            {
              int j;
              rsbac_acl_rights_vector_t tmp_rv;

              for(j=0; j<ACLR_NONE; j++)
                if(   (argv[1][j] != '0')
                   && (argv[1][j] != '1')
                  )
                  {
                    fprintf(stderr, "%s: Invalid mask vector %s\n", progname, argv[1]);
                    exit(1);
                  }
              strtou64acl(argv[1], &tmp_rv);
              rights_vector |= tmp_rv;
              argv++;
              argc--;
              continue;
            }
          right = get_acl_special_right_nr(argv[1]);
          if((right == R_NONE) || (right == ACLR_NONE))
            {
              right = strtol(argv[1],0,10);
              if(   (   (right >= R_NONE)
                     && (right < RSBAC_ACL_SPECIAL_RIGHT_BASE)
                    )
                 || (right >= ACLR_NONE)
                 || (   (right == 0)
                     && strcmp(argv[1],"0")
                    )
                )
                {
                  if(!strcmp(argv[1],"RW"))
                    {
                      rights_vector |= RSBAC_READ_WRITE_REQUEST_VECTOR;
                    }
                  else
                  if(!strcmp(argv[1],"SY"))
                    {
                      rights_vector |= RSBAC_SYSTEM_REQUEST_VECTOR;
                    }
                  else
                  if(!strcmp(argv[1],"SE"))
                    {
                      rights_vector |= RSBAC_SECURITY_REQUEST_VECTOR;
                    }
                  else
                  if(!strcmp(argv[1],"S"))
                    {
                      rights_vector |= RSBAC_ACL_SPECIAL_RIGHTS_VECTOR;
                    }
                  else
                  if(!strcmp(argv[1],"R"))
                    {
                      rights_vector |= RSBAC_READ_REQUEST_VECTOR;
                      rused = TRUE;
                    }
                  else
                  if(!strcmp(argv[1],"W"))
                    {
                      rights_vector |= RSBAC_WRITE_REQUEST_VECTOR;
                      wused = TRUE;
                    }
                  else
                  if(!strcmp(argv[1],"A"))
                    {
                      rights_vector |= RSBAC_ALL_REQUEST_VECTOR;
                    }
                  else
                  if(!strcmp(argv[1],"NWS"))
                    {
                      rights_vector |= RSBAC_NWS_REQUEST_VECTOR;
                    }
                  else
                  if(!strcmp(argv[1],"NWR"))
                    {
                      rights_vector |= RSBAC_NWR_REQUEST_VECTOR;
                      rused = TRUE;
                    }
                  else
                  if(!strcmp(argv[1],"NWW"))
                    {
                      rights_vector |= RSBAC_NWW_REQUEST_VECTOR;
                      wused = TRUE;
                    }
                  else
                  if(!strcmp(argv[1],"NWC"))
                    {
                      rights_vector |= RSBAC_NWC_REQUEST_VECTOR;
                    }
                  else
                  if(!strcmp(argv[1],"NWE"))
                    {
                      rights_vector |= RSBAC_NWE_REQUEST_VECTOR;
                    }
                  else
                  if(!strcmp(argv[1],"NWA"))
                    {
                      rights_vector |= RSBAC_NWA_REQUEST_VECTOR;
                    }
                  else
                  if(!strcmp(argv[1],"NWF"))
                    {
                      rights_vector |= RSBAC_NWF_REQUEST_VECTOR;
                    }
                  else
                  if(!strcmp(argv[1],"NWM"))
                    {
                      rights_vector |= RSBAC_NWM_REQUEST_VECTOR;
                    }
                  else
                    { /* end of rights */
                      break;
                    }
                }
            }
          else
            {
              rights_vector |= ((rsbac_acl_rights_vector_t) 1 << right);
            }
          argv++;
          argc--;
        }
      if(rused && wused)
        {
          rights_vector |= RSBAC_READ_WRITE_OPEN_REQUEST_VECTOR;
        }

      target = get_target_nr(argv[1]);
      /* trim rights_vector for target */
      switch(target)
        {
          case T_DIR:
          case T_FILE:
          case T_FD:
            argv++;
            argc--;
            rights_vector &= (RSBAC_ALL_REQUEST_VECTOR | RSBAC_ACL_SPECIAL_RIGHTS_VECTOR);
            break;
          case T_DEV:
            argv++;
            argc--;
            rights_vector &= (RSBAC_ALL_REQUEST_VECTOR | RSBAC_ACL_SPECIAL_RIGHTS_VECTOR);
            break;
          case T_IPC:
            argv++;
            argc--;
            rights_vector &= (RSBAC_IPC_REQUEST_VECTOR | RSBAC_ACL_SPECIAL_RIGHTS_VECTOR);
            break;
          case T_SCD:
            argv++;
            argc--;
            rights_vector &= (RSBAC_SCD_REQUEST_VECTOR | RSBAC_NONE_REQUEST_VECTOR | RSBAC_ACL_SPECIAL_RIGHTS_VECTOR);
            break;
          case T_USER:
            argv++;
            argc--;
            rights_vector &= (RSBAC_ACL_USER_RIGHTS_VECTOR | RSBAC_ACL_SPECIAL_RIGHTS_VECTOR);
            break;
          case T_PROCESS:
            argv++;
            argc--;
            rights_vector &= (RSBAC_PROCESS_REQUEST_VECTOR | RSBAC_ACL_SPECIAL_RIGHTS_VECTOR);
            break;
          case T_NONE:
            fprintf(stderr, "%s: No target type given, assuming FD\n", progname);
            target = T_FD;
            rights_vector &= (RSBAC_ALL_REQUEST_VECTOR | RSBAC_ACL_SPECIAL_RIGHTS_VECTOR);
            break;
          default:
            fprintf(stderr, "%s: Invalid target type %s\n", progname, argv[1]);
            exit(1);
        }
      if(verbose)
        {
          char tmp1[120];

          if(set)
            {
              printf("Set mask: %s\n",
                     u64tostracl(tmp1, rights_vector));
              if(printall)
              {
                int i;

                for (i=0; i<R_NONE; i++)
                  if(rights_vector & ((rsbac_acl_rights_vector_t) 1 << i))
                    printf("  %s\n", get_request_name(tmp1,i));
                for (i=RSBAC_ACL_SPECIAL_RIGHT_BASE; i<ACLR_NONE; i++)
                  if(rights_vector & ((rsbac_acl_rights_vector_t) 1 << i))
                    printf("  %s\n", get_acl_special_right_name(tmp1,i));
              }
            }
          else
            {
              printf("Get mask.\n");
            }
        }
      if(verbose>1)
        printf("\n%s: %i targets\n\n", progname, argc - 1);
      for (i=1;i < (argc);i++)
        {
          process(argv[i]);
        }
    }
  else
    {
      printf("%s (RSBAC %s)\n***\n", progname, VERSION);
      printf("Use: %s [switches] [rights] target-type file/dirname(s)\n", progname);  
      printf("- -v = verbose, -r = recurse into subdirs,\n");
      printf("- -p = print right names, -s = set mask, not get\n");
      printf("- -b = backup mode, -n = list valid SCD names\n");
      printf("  rights = list of space-separated right names (requests and ACL specials),\n");
      printf("        also request groups R (read requests), RW (read-write),\n");
      printf("        SY (system), SE (security), A (all)\n");
      printf("        S (ACL special rights)\n");
      printf("        and NWx with x = S R W C E A F M (similar to well-known network system)\n");
      printf("- target-type = FILE, DIR, DEV, IPC, SCD, USER, PROCESS or FD\n");
      printf("  (FD: let %s decide between FILE and DIR, no DEV),\n", progname);
      printf("  (IPC, USER, PROCESS: only :DEFAULT:\n");
      printf("- Use name :DEFAULT: for default ACL\n");
    }
  return (res);
}

