#include <stdio.h>
#include <stdlib.h>
#include <pwd.h>
#include <sys/types.h>
#include <rsbac/types.h>
#include <rsbac/aci_data_structures.h>
#include <rsbac/getname.h>
#include <rsbac/syscalls.h>
#include <rsbac/error.h>
#include <rsbac/helpers.h>

int main(int argc, char ** argv)
{
  int attr_list[RSBAC_USER_NR_ATTRIBUTES] = RSBAC_USER_ATTR_LIST;
  int res = 0;
  u_int position;
  char tmp1[80],tmp2[120];
  int j;
  union rsbac_attribute_value_t value;
  union rsbac_target_id_t tid;
  enum rsbac_attribute_t attr;
  struct passwd * user_info_p;
  int inherit = 0;
  int numeric = 0;
  int both = 0;
  int printall = 0;
  char * progname = argv[0];
  boolean use_ori = FALSE;

  while((argc > 1) && (argv[1][0] == '-'))
    {
      char * pos = argv[1];
      pos++;
      while(*pos)
        {
          switch(*pos)
            {
              case 'e':
                inherit=1;
                break;
              case 'p':
                printall=1;
                break;
              case 'n':
                numeric=1;
                break;
              case 'b':
                both=1;
                break;
              case 'l':
                while((user_info_p = getpwent()))
                  {
                    if(numeric)
                      printf("%u\n", user_info_p->pw_uid);
                    else
                    if(both)
                      printf("%s %u\n", user_info_p->pw_name, user_info_p->pw_uid);
                    else
                      printf("%s\n", user_info_p->pw_name);
                  }
                exit(0);

              default:
                fprintf(stderr, "%s: unknown parameter %c\n", progname, *pos);
            }
          pos++;
        }
      argv++;
      argc--;
    }
  switch(argc)
    {
      case 3:
        value.dummy = -1;
        if(   (!(user_info_p = getpwnam(argv[1])))
           && (   (!(user_info_p = getpwuid(strtol(argv[1],0,0))))
               || (!(user_info_p->pw_uid) && strcmp("0",argv[1])) )
          )
            {
              use_ori = TRUE;
              tid.user = strtol(argv[1],0,0);
            }
        else
          tid.user = user_info_p->pw_uid;
        if(!strcmp("user_nr",argv[2]))
          {
            if(use_ori)
              printf("%s\n", argv[1]);
            else
              printf("%i\n", user_info_p->pw_uid);
            exit(0);
          }
        if(!strcmp("user_name",argv[2]))
          {
            if(use_ori)
              printf("%s\n", argv[1]);
            else
              printf("%s\n", user_info_p->pw_name);
            exit(0);
          }
        if(!strcmp("full_name",argv[2]))
          {
            if(use_ori)
              printf("%s\n", argv[1]);
            else
              printf("%s\n", user_info_p->pw_gecos);
            exit(0);
          }

        attr = get_attribute_nr(argv[2]);
        if(attr == A_none)
          {
            fprintf(stderr, "%s: Invalid Attribute %s!\n", progname, argv[2]);
            exit(1);
          }

        res = rsbac_get_attr(T_USER, &tid, attr, &value, inherit);
        error_exit(res);
        switch(attr)
          {
            case A_rc_type:
            case A_rc_type_fd:
            case A_rc_force_role:
            case A_rc_role:
            case A_rc_def_role:
            case A_security_level:
              printf("%u\n",value.u_char_dummy);
              break;
            case A_mac_categories:
              printf("%s\n",u64tostrmac(tmp1,value.mac_categories));
              break;
            case A_log_user_based:
              if(printall)
                {
                  int i;

                  for (i=0; i<R_NONE; i++)
                    if(value.log_user_based & ((rsbac_request_vector_t) 1 << i))
                      printf(" %s\n", get_request_name(tmp1,i));
                }
              else
                printf("%s\n",u64tostrlog(tmp1,value.log_user_based));
              break;
            default:
              printf("%i\n",value.dummy);
          }
        exit(0);

      case 4:
        attr = get_attribute_nr(argv[2]);
        if(   (!(user_info_p = getpwnam(argv[1])))
           && (   (!(user_info_p = getpwuid(strtol(argv[1],0,0))))
               || (!(user_info_p->pw_uid) && strcmp("0",argv[1])) )
              )
          {
            fprintf(stderr, "Invalid user %s!\n", argv[1]);
            exit(1);
          }
        tid.user = user_info_p->pw_uid;
        switch(attr)
          {
            case A_mac_categories:
              position = strtol(argv[3],0,10);
              if(position > RSBAC_MAC_MAX_CAT)
                {
                  fprintf(stderr, "Invalid position counter %s\n", argv[3]);
                  exit(1);
                }
              res = rsbac_get_attr(T_USER, &tid, A_mac_categories, &value, inherit);
              error_exit(res);
              printf("%u\n",
                     (u_int) (value.mac_categories >> position) & 1);
              exit(0);

            case A_log_user_based:
              position = get_request_nr(argv[3]);
              if(position >= R_NONE)
                {
                  fprintf(stderr, "Invalid request %s\n", argv[3]);
                  exit(1);
                  }
                res = rsbac_get_attr(T_USER, &tid, A_log_user_based, &value, inherit);
                error_exit(res);
                printf("%u\n",
                     (u_int) (value.log_program_based >> position) & 1);
                exit(0);
            default:
          }

      default:
        printf("%s (RSBAC %s)\n***\n", progname, VERSION);
        printf("Use: %s [switches] user attribute [position|request-name]\n\n", progname);  
        printf("- -e = show effective (maybe inherited) value, not real\n");
        printf("- -n = numeric value, -b = both names and numbers, -l list all users\n");
        printf("- attribute (string) and returned value = see following list:\n");
        printf("mac_categories\t\t(with additional parameter position)\n\t\t\t0=no, 1=yes\n");
        printf("log_user_based\t(with additional parameter request-name)\n\t\t\t0=no, 1=yes\n");
        for (j=0;j<RSBAC_USER_NR_ATTRIBUTES;j++)
          {
            get_attribute_name(tmp1,attr_list[j]);
            get_attribute_param(tmp2,attr_list[j]);
            printf("%s \t%s\n",tmp1,tmp2);
          }
    }
  exit(1);
}
