/*Simple daemon logger for RSBAC
 ***************************************************
  You must change capabilities list for this program 
  to grand operations with RSBAC log
 ***************************************************
  rklog (c) - Stanislav I. Ievlev, 2000
  RSBAC (c) - Amon Ott., 1995-2000
  Some part of code from klogd - (c) 1995  Martin Schulze <Martin.Schulze@Linux.DE>
  Some changes made by Amon Ott, 2000
  */

/* Includes. */
#include <unistd.h>
#include <signal.h>
#include <errno.h>
#include <sys/fcntl.h>
#include <sys/stat.h>
#include <time.h>
#include <stdarg.h>
#include <paths.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <syslog.h>

#include <rsbac/syscalls.h>
#include "pidfile.h"

#ifndef _PATH_RSBAC_LOG
#define _PATH_RSBAC_LOG "/proc/rsbac-info/rmsg"
#endif

#ifndef _PATH_RSBAC_OUT
#define _PATH_RSBAC_OUT "/tmp/security-out"
#endif

#ifndef _PATH_PIDFILE
#define _PATH_PIDFILE "/var/run/rklogd.pid"
#endif

#ifndef _SECOFFUID
#define _SECOFFUID 400
#endif

#define LOG_LINE_SIZE 4096
#define LOG_LINE_LENGTH 1000

/*define for testing*/
//#define TEST 
#define MAX_NAME 200

int	rsbac_msg  ; /*message file*/
FILE	*rsbac_out ; /*log file*/
char    rsbac_string[LOG_LINE_SIZE];

int    use_proc=1;

extern int errno;
void stop_work(int);
void write_message_log();


void init_files(char *log_filename)
{
	/*open message file*/
    if (use_proc) /*use proc*/{
	if ( (rsbac_msg = open(_PATH_RSBAC_LOG, O_RDONLY)) < 0 )
	{
		fprintf(stderr, "rklogd: Cannot open proc file system , " \
			"%d - %s.\n", errno, strerror(errno));
			
		exit(1);
	};
    }else{ /*use kernel*/
	if (rsbac_log(1,rsbac_string,sizeof(rsbac_string)) <0)
	{
		fprintf(stderr, "rklogd: Cannot open kernel log buffer , " \
			"%d - %s.\n", errno, strerror(errno));
	               
	    exit(1);
	};
    }
    
	
	/*open log file*/
	if ( (rsbac_out = fopen(log_filename, "a")) < 0 )
	{
		fprintf(stderr, "rklogd: Cannot open log file , " \
			"%d - %s.\n", errno, strerror(errno));
			
		exit(1);
	}
	chmod(log_filename,0666);
        
}

void read_message_kernel()
{
auto int rdcnt;

	/*
	 * Zero-fill the log buffer.  This should cure a multitude of
	 * problems with klogd logging the tail end of the message buffer
	 * which will contain old messages.  Then read the kernel log
	 * messages into this fresh buffer.
	 */
	memset(rsbac_string, '\0', sizeof(rsbac_string));
	if ( (rdcnt = rsbac_log(2, rsbac_string, sizeof(rsbac_string)-1)) < 0 )
	{
		if ( errno == EINTR )
			return;
#ifdef TEST
	fprintf(stderr, "rklogd: Cannot read from kernel , " \
                 			"%d - %s.\n", errno, strerror(errno));
#endif
	syslog(LOG_INFO, "rklogd: Can't read from kernel.");
	}
	
#ifdef TEST	
        fprintf(stderr,"Receive string from kernel: %s",rsbac_string);
#endif

	return;

}
void read_message_proc()
{
auto int rdcnt;

/*
* Zero-fill the log buffer.  This should cure a multitude of
* problems with klogd logging the tail end of the message buffer
* which will contain old messages.  Then read the kernel messages
* from the message pseudo-file into this fresh buffer.
*/
	memset(rsbac_string, '\0', sizeof(rsbac_string));
	if ( (rdcnt = read(rsbac_msg, rsbac_string, sizeof(rsbac_string)-1)) < 0 )
	{
		if ( errno == EINTR )
			return;
#ifdef TEST
	fprintf(stderr, "rklogd: Cannot read from proc , " \
                 			"%d - %s.\n", errno, strerror(errno));
#endif
	syslog(LOG_INFO, "rklogd: Can't read from proc file system.");
	}

#ifdef TEST	
        fprintf(stderr,"Receive string from proc: %s",rsbac_string);
#endif

       return;
}

void stop_work(int sig)
{
        syslog(LOG_INFO,"security kernel logger stopped");
	
	fclose(rsbac_out);
	if (use_proc)
          close(rsbac_msg); /*close proc*/
	else
	  rsbac_log(0, rsbac_string, sizeof(rsbac_string) ); /*close kernel log*/
	/*close system log*/
	closelog();
#ifdef TEST
        fprintf(stderr,"exiting...");
#endif

        remove_pid(_PATH_PIDFILE);
	
	exit (0);
}

void write_message_log()
{
char tmpstr[LOG_LINE_SIZE+20];
time_t now;
char timestamp[20];
char *workstr=rsbac_string,*workstr1;
int length,new_length;
/*make time stamp*/
now=time(0);
strncpy(timestamp,ctime(&now),19);

/*split message into separate lines*/

length=strlen(workstr);
while (length){
/*point to future line*/
workstr1=strchr(workstr,'\n');
workstr1++;
new_length=strlen(workstr1);
strncpy(tmpstr,workstr,(length-new_length-1));
tmpstr[length-new_length-1]='\0';

length=new_length;
workstr=workstr1;

fprintf(rsbac_out,"%s : %s\n",timestamp,tmpstr);
}

/*flush file stream*/
fflush(rsbac_out);

}



int main(int argc,char *argv[])
{
int i,ch;
uid_t secoffuid = _SECOFFUID;
char *logname=malloc(MAX_NAME);



/*open central system log for bug errors*/
openlog("rklogd",LOG_PID,LOG_AUTH);

/*analize options*/
strncpy(logname,_PATH_RSBAC_OUT,MAX_NAME);
while ((ch = getopt(argc, argv, "spf:u:")) != EOF)
    switch((char)ch)
	{
	    case 's':		/* Use kernel syscall*/
		use_proc = 0;
		break;
	    case 'p':		/* Use proc */
		use_proc = 1;
		break;
	    case 'f':		/* Define an output file. */
		logname = optarg;
		break;
	    case 'u':		/* Define an output file. */
		secoffuid = strtol(optarg,0,0);
		break;
	}
#ifdef TEST
if (use_proc)
 fprintf(stderr,"read from proc\n");
else
 fprintf(stderr,"read from kernel\n");
 
 fprintf(stderr,"log file name:%s\n",logname);
#endif
/*signals handlers*/
#ifndef TEST
for (i = 1; i < NSIG; i++)	signal(i, SIG_IGN); //ignore all signals.

signal(SIGINT, stop_work); 
signal(SIGKILL, stop_work);
signal(SIGTERM, stop_work);

/*make a daemon*/
if ( fork() == 0 )
    {
	/*we need stay in root for create pid file*/	 
	if (!check_pid(_PATH_PIDFILE))
	{
		if (!write_pid(_PATH_PIDFILE))
			stop_work(9);
	}
	else
	{
		fputs("rklogd: Already running.\n", stderr);
		stop_work(9);
	}

      /*change uid to secoff*/
      if (setuid(secoffuid)<0){
      fprintf(stderr,"rklogd:Cannot change UID to %u\n", secoffuid);
      exit(1);
      }
      /* This is the child closing its file descriptors. */
      
      /*we need secoff for do it*/
      init_files(logname);
      /*free(logname);*/
	
      fclose(stdout);
      fclose(stderr);
      fclose(stdin);

      setsid();
    }else /*parent process*/
    exit(0);
#endif
/*report to system log*/
syslog(LOG_INFO,"security kernel logger started.",32);
/*go to infinity loop*/
while (1) {
/* read next line from _______(kernel, proc) */
switch (use_proc){
case 1:    read_message_proc();
	   break;
case 2:    read_message_kernel();
	   break;
case 3:    pause();
	   break;
}/* switch() */
write_message_log();
}

syslog(LOG_INFO,"strange bug.",14);
closelog();

return 0;
}

